#-*- coding: utf-8 -*-

import re
import time
from datetime import datetime

from django.core.paginator import InvalidPage
from django.utils import timezone
from django.conf import settings

__all__ = [
    'page_key',
    'to_page_key',
    'InvalidPage']

PAGE_RE = re.compile(r'^(?P<value>[0-9]+\.[0-9]{6})-(?P<pk>[0-9]+)$')


def _make_aware_maybe(dt):
    if not getattr(settings, 'USE_TZ', False):
        return dt
    if timezone.is_aware(dt):
        return dt.astimezone(timezone.utc)
    return timezone.make_aware(dt, timezone=timezone.utc)


def _fromtimestamp(ts):
    if not getattr(settings, 'USE_TZ', False):
        return datetime.fromtimestamp(ts)
    return datetime.utcfromtimestamp(ts)


# py2 only
def _timestamp(dt):
    if timezone.is_naive(dt):
        return time.mktime((
            dt.year, dt.month, dt.day,
            dt.hour, dt.minute, dt.second,
            -1, -1, -1)) + dt.microsecond / 1e6
    return (dt - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds()


def page_key(raw_page):
    """
    Parse a raw page value of ``timestamp-pk`` format.
    Return a tuple of (timestamp, pk)

    :raises: ``InvalidPage``
    """
    if not raw_page:
        return None, None
    m = re.match(PAGE_RE, raw_page)
    if m is None:
        raise InvalidPage('Bad key format')
    try:
        timestamp = _fromtimestamp(float(m.group('value')))
    except (OverflowError, ValueError):
        raise InvalidPage('Key out of range')
    return _make_aware_maybe(timestamp), m.group('pk')


def to_page_key(value=None, pk=None):
    """Serialize a value and pk to `timestamp-pk`` format"""
    if value is None:
        return ''
    value = _make_aware_maybe(value)
    try:
        timestamp = value.timestamp()
    except AttributeError:
        timestamp = _timestamp(value)
    return '{:.6f}-{}'.format(timestamp, pk)



#-*- coding: utf-8 -*-


try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

from django.core.paginator import EmptyPage
from django.db.models import QuerySet

__all__ = [
    'SeekPaginator',
    'SeekPage',
    'EmptyPage',
    'NEXT_PAGE',
    'PREV_PAGE']


NEXT_PAGE = 1
PREV_PAGE = 2


class _NoPk:
    def __str__(self):
        return 'NoPk'

    def __repr__(self):
        return '<NoPk>'

# We need a distinct value for None,
# because None is a valid pk when fetching
# the first page
_NO_PK = _NoPk()


class SeekPaginator(object):

    def __init__(self, query_set, per_page, lookup_field):
        assert isinstance(query_set, QuerySet), 'QuerySet expected'
        assert isinstance(per_page, int), 'Int expected'
        assert isinstance(lookup_field, str), 'String expected'
        self.query_set = query_set
        self.per_page = per_page
        self.is_desc = lookup_field.startswith('-')
        self.is_asc = not self.is_desc
        self.lookup_field = lookup_field.lstrip('-')

    def prepare_order(self, has_pk, move_to):
        pk_sort = 'pk'
        lookup_sort = self.lookup_field
        if ((self.is_desc and move_to == NEXT_PAGE) or
                (self.is_asc and move_to == PREV_PAGE)):
            pk_sort = '-%s' % pk_sort
            lookup_sort = '-%s' % lookup_sort
        if has_pk:
            return [lookup_sort, pk_sort]
        return [lookup_sort]

    def prepare_lookup(self, value, pk, move_to):
        lookup_include = '%s__gt' % self.lookup_field
        lookup_exclude_pk = 'pk__lte'
        if ((self.is_desc and move_to == NEXT_PAGE) or
                (self.is_asc and move_to == PREV_PAGE)):
            lookup_include = '%s__lt' % self.lookup_field
            lookup_exclude_pk = 'pk__gte'
        lookup_exclude = None
        if pk is not _NO_PK:
            lookup_include = "%se" % lookup_include
            lookup_exclude = {self.lookup_field: value, lookup_exclude_pk: pk}
        lookup_filter = {lookup_include: value}
        return lookup_filter, lookup_exclude

    def apply_filter(self, value, pk, move_to):
        query_set = self.query_set
        lookup_filter, lookup_exclude = self.prepare_lookup(
            value=value, pk=pk, move_to=move_to)
        query_set = query_set.filter(**lookup_filter)
        if lookup_exclude:
            query_set = query_set.exclude(**lookup_exclude)
        return query_set

    def seek(self, value, pk, move_to):
        """
        Skip the current record in case
        there are multiple with the same value

        Lookup next page (-DESC)::

            # ...
            WHERE date <= ?
            AND NOT (date = ? AND id >= ?)
            ORDER BY date DESC, id DESC

        """
        query_set = self.query_set
        if value is not None:
            query_set = self.apply_filter(
                value=value, pk=pk, move_to=move_to)
        query_set = query_set.order_by(
            *self.prepare_order(
                has_pk=pk is not _NO_PK, move_to=move_to))
        return query_set

    def page(self, value, pk=_NO_PK, move_to=NEXT_PAGE):
        """
        The param ``value`` may be ``None`` to get the first page.
        Pass both ``value`` and ``pk`` when the ``lookup_field``'s model
        field is not unique. Otherwise, just pass the ``value``

        :raises: ``EmptyPage``
        """
        assert move_to in (NEXT_PAGE, PREV_PAGE)
        query_set = self.seek(
            value=value, pk=pk, move_to=move_to)

        if value and not query_set.exists():
            raise EmptyPage()

        return SeekPage(
            query_set=query_set,
            key={'value': value, 'pk': pk},
            move_to=move_to,
            paginator=self)


class SeekPage(Sequence):

    def __init__(self, query_set, key, move_to, paginator):
        self._query_set = query_set
        self._key = key
        self._move_to = move_to
        self._object_list = None
        self.paginator = paginator

    def __repr__(self):
        return '<Page value={value} pk={pk}>'.format(**self._key)

    def __len__(self):
        return len(self.object_list)

    def __getitem__(self, index):
        return self.object_list[index]

    @property
    def object_list(self):
        if self._object_list is not None:
            return self._object_list
        # We could fetch an extra item an
        # save a query on has_prev/next_page,
        # but meh, the query for hax_X should be fairly cheap
        self._object_list = list(self._query_set[:self.paginator.per_page])
        if self._move_to == PREV_PAGE:
            self._object_list.reverse()
        return self._object_list

    def _some_seek(self, direction):
        assert self.object_list
        assert direction in (NEXT_PAGE, PREV_PAGE)
        last = self.object_list[0]
        if direction == NEXT_PAGE:
            last = self.object_list[-1]
        pk = _NO_PK
        if self._key['pk'] is not _NO_PK:
            pk = last.pk
        return self.paginator.seek(
            value=getattr(last, self.paginator.lookup_field),
            pk=pk,
            move_to=direction)

    def has_next(self):
        if not self.object_list:
            return False
        return self._some_seek(NEXT_PAGE).exists()

    def has_previous(self):
        if not self.object_list:
            return False
        return self._some_seek(PREV_PAGE).exists()

    def next_objects_left(self, limit=None):
        """Return the number of next records"""
        if not self.object_list:
            return 0
        qs = self._some_seek(NEXT_PAGE)
        if limit:
            qs = qs[:limit]
        return qs.count()

    def prev_objects_left(self, limit=None):
        """Return the number of prev records"""
        if not self.object_list:
            return 0
        qs = self._some_seek(PREV_PAGE)
        if limit:
            qs = qs[:limit]
        return qs.count()

    def _some_pages_left(self, direction, limit):
        some_objects_left = self.prev_objects_left
        if direction == NEXT_PAGE:
            some_objects_left = self.next_objects_left
        limit = (limit or 0) * self.paginator.per_page
        return (-some_objects_left(limit) // self.paginator.per_page) * -1  # ceil

    def next_pages_left(self, limit=None):
        """Return the number of next pages"""
        return self._some_pages_left(NEXT_PAGE, limit)

    def prev_pages_left(self, limit=None):
        """Return the number of prev pages"""
        return self._some_pages_left(PREV_PAGE, limit)

    def _some_page(self, index):
        if not self.object_list:
            return {}
        key = {
            'value': getattr(
                self.object_list[index],
                self.paginator.lookup_field)}
        if self._key['pk'] is not _NO_PK:
            key['pk'] = self.object_list[index].pk
        return key

    def next_page(self):
        """Return ``{'value': value, 'pk' pk}`` to fetch the next page"""
        return self._some_page(-1)

    def prev_page(self):
        """Return ``{'value': value, 'pk' pk}`` to fetch the prev page"""
        return self._some_page(0)


def paginate(query_set, per_page, lookup_field, value, pk=None, move_to=NEXT_PAGE):
    """Return a ``SeekPage`` containing the paginated result"""
    return (
        SeekPaginator(
            query_set=query_set,
            per_page=per_page,
            lookup_field=lookup_field)
        .page(
            value=value,
            pk=pk,
            move_to=move_to))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup


URL = 'https://github.com/nitely/django-infinite-scroll-pagination'
README = "For more info, go to: {}".format(URL)

VERSION = __import__('infinite_scroll_pagination').__version__

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-infinite-scroll-pagination',
    version=VERSION,
    description=(
        'infinite-scroll-pagination is a Django lib that implements'
        'the *seek method* for scalable pagination.'),
    author='Esteban Castro Borsani',
    author_email='ecastroborsani@gmail.com',
    long_description=README,
    url=URL,
    packages=[
        'infinite_scroll_pagination',
    ],
    include_package_data=True,
    zip_safe=False,
    license='MIT License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
